

/* stattree.c */


/*------------------------------------------------------------------------
 * Solitaire field statistics: Tree
 *------------------------------------------------------------------------
 * This version of stattree is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#include "stattree.h"
#include "mymemory.h"
#include "wordcnt.h"
#include "field.h"

#include <assert.h>

/*----------------------------------------------------------------------*/


struct s_STATTREE
{
    tpsWORDCNT psPatterns;   /* (1)          */
    tpsWORDCNT psMovesPat;   /* (MovesRange) */
    tpsWORDCNT psMovesSum;   /* (1)          */
    tpsWORDCNT psWaysEnd;    /* (1)          */
};

typedef struct s_STATTREE tsSTATTREE;


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Create a new STATTREE
 *-----------------------------------------------
 */

tpsSTATTREE psStatTreeCreate (tVOID)
{
    tpsSTATTREE psTree;

    psTree = pvMemoryAllocObject(sSTATTREE);

    psTree->psPatterns = psWordCntCreate(1);
    psTree->psMovesPat = psWordCntCreate(cnFieldMovesRange);
    psTree->psMovesSum = psWordCntCreate(1);
    psTree->psWaysEnd  = psWordCntCreate(1);

    vWordCntInit0All(psTree->psPatterns);
    vWordCntInit0All(psTree->psMovesPat);
    vWordCntInit0All(psTree->psMovesSum);
    vWordCntInit0All(psTree->psWaysEnd );

    return (psTree);
}


/*-----------------------------------------------
 * Destroy a STATTREE
 *-----------------------------------------------
 */

tpsSTATTREE psStatTreeDestroy (tpsSTATTREE psTree)
{
    if (psTree)
      {
        psWordCntDestroy(psTree->psPatterns);
        psWordCntDestroy(psTree->psMovesPat);
        psWordCntDestroy(psTree->psMovesSum);
        psWordCntDestroy(psTree->psWaysEnd );

        pvMemoryFree(psTree);
      }

    return (NULL);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsSTATTREE psStatTreeMemUsage (tpsSTATTREE psTree, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (psTree)
      {
        psWordCntMemUsage(psTree->psPatterns,pcnSizeof,pcnBlocks);
        psWordCntMemUsage(psTree->psMovesPat,pcnSizeof,pcnBlocks);
        psWordCntMemUsage(psTree->psMovesSum,pcnSizeof,pcnBlocks);
        psWordCntMemUsage(psTree->psWaysEnd ,pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sSTATTREE);
        *pcnBlocks += 1;
      }

    return (psTree);
}


/*-----------------------------------------------
 * Add a field statistic
 *-----------------------------------------------
 */

tVOID vStatTreeAddField (tpsSTATTREE psTree, tpsFIELDSTT psStt, tBOOL bTreeAll)
{
    tCOUNT cnSymType,cnMoves;

    cnSymType = ( bTreeAll ? cnFieldSttGetSymType(psStt) : 1 );
    cnMoves   = ( bTreeAll ? cnFieldSttGetMovesAll(psStt) : cnFieldSttGetMovesNrm(psStt) );

    vWordCntAddWord(psTree->psPatterns,0,(tUWORD)cnSymType);
    vWordCntAddWord(psTree->psMovesPat,cnMoves,(tUWORD)cnSymType);

    if (cnMoves == 0)
      {
        if (bTreeAll)
            vFieldSttAddCountAll(psStt,psTree->psWaysEnd,0);
        else
            vFieldSttAddCountNrm(psStt,psTree->psWaysEnd,0);
      }
}


/*-----------------------------------------------
 * Add a STATTREE to another STATTREE
 *-----------------------------------------------
 * The two objects are expected to be from the
 * same tree, else the statistic would be useless.
 *-----------------------------------------------
 */

tVOID vStatTreeAddObject (tpsSTATTREE psTree, tpsSTATTREE psAdd)
{
    tCOUNT cnI;

    vWordCntAdd(psTree->psPatterns,0,psAdd->psPatterns,0);

    for (cnI = 0; cnI < cnFieldMovesRange; cnI++)
        vWordCntAdd(psTree->psMovesPat,cnI,psAdd->psMovesPat,cnI);

    vWordCntAdd(psTree->psMovesSum,0,psAdd->psMovesSum,0);
    vWordCntAdd(psTree->psWaysEnd ,0,psAdd->psWaysEnd ,0);
}


/*-----------------------------------------------
 * Calculate statistic
 *-----------------------------------------------
 */

tVOID vStatTreeCalcWhole (tpsSTATTREE psTree)
{
    tCOUNT cnI;

    for (cnI = 0; cnI < cnFieldMovesRange; cnI++)
        vWordCntAddMulti(psTree->psMovesSum,0,psTree->psMovesPat,cnI,(tUWORD)cnI);
}


/*-----------------------------------------------
 * Test if there is any pattern at all
 *-----------------------------------------------
 */

tBOOL bStatTreeIsEmpty (tpsSTATTREE psTree)
{
    return ( bWordCntIs0(psTree->psPatterns,0) );
}


/*-----------------------------------------------
 * Get the maximum number of moves
 *-----------------------------------------------
 */

tCOUNT cnStatTreeMaxMoves (tpsSTATTREE psTree)
{
    tCOUNT cnI,cnMax = 0;

    for (cnI = 0; cnI < cnFieldMovesRange; cnI++)
        if ( !bWordCntIs0(psTree->psMovesPat,cnI) )
            cnMax = cnI;

    return (cnMax);
}


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Store parameter into table
 *-----------------------------------------------
 * NULL is not valid for pzContent.
 * Content qualifier:
 * "P" : Patterns
 * "M" : MovesPat[cnMove]
 * "Mn": MovesPat[n] with n = decimal number
 *       MovesSum if cnMove or n >= MovesRange
 * "S" : MovesSum
 * "W" : WaysEnd
 * Any value can have a subqualifier for the
 * SetCellWordCnt format qualifier.
 *-----------------------------------------------
 */

tVOID vStatTreeTabSet (tpsSTATTREE psTree, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tCOUNT     cnIndex;
    tpsWORDCNT psCnt = NULL;

    switch (*pzContent)
      {
      case 'P' :
        psCnt   = psTree->psPatterns;
        cnIndex = 0;
        pzContent++;
        break;
      case 'S' :
        psCnt   = psTree->psMovesSum;
        cnIndex = 0;
        pzContent++;
        break;
      case 'W' :
        psCnt   = psTree->psWaysEnd;
        cnIndex = 0;
        pzContent++;
        break;

      case 'M' :
        psCnt     = psTree->psMovesPat;
        cnIndex   = cnStatTabsGetMove(psTabs);
        pzContent = pzStatTabsStrGetNumber(pzContent + 1,&cnIndex);
        if (cnIndex >= cnFieldMovesRange)
          {
            psCnt   = psTree->psMovesSum;
            cnIndex = 0;
          }
        break;

      default : assert(FALSE);
      }

    if (psCnt)
        vStatTabsSetCellWordCnt(psTabs,psCnt,cnIndex,*pzContent);
}


/*----------------------------------------------------------------------*/
