

/* tabcell.c */


/*------------------------------------------------------------------------
 * Module for HTML tables: single cell
 *------------------------------------------------------------------------
 * This version of tabcell is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#include "tabcell.h"
#include "mymemory.h"
#include "mystring.h"

/*----------------------------------------------------------------------*/


struct s_TABCELL
{
    tPSZ   pzText;

    tCOUNT cnColSpan;
    tCOUNT cnRowSpan;

    tULONG ulColorBack;  /* 0x00rrggbb; 0xnn------ == default */
    tULONG ulColorFont;

    tCHAR  cAlign;
    tBOOL  bBroad;
    tBOOL  bItalic;
};

typedef struct s_TABCELL tsTABCELL;


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Create a new TABCELL
 *-----------------------------------------------
 */

tpsTABCELL psTabCellCreate (tVOID)
{
    tpsTABCELL psCell;

    psCell = pvMemoryAllocObject(sTABCELL);

    psCell->pzText = NULL;

    psCell->cnColSpan =
    psCell->cnRowSpan = 1;

    psCell->ulColorBack =
    psCell->ulColorFont = 0xff000000;

    psCell->cAlign  = 'c';
    psCell->bBroad  =
    psCell->bItalic = FALSE;

    return (psCell);
}


/*-----------------------------------------------
 * Destroy a TABCELL
 *-----------------------------------------------
 */

tpsTABCELL psTabCellDestroy (tpsTABCELL psCell)
{
    if (psCell)
      {
        ppzStringClear(&(psCell->pzText));

        pvMemoryFree(psCell);
      }

    return (NULL);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsTABCELL psTabCellMemUsage (tpsTABCELL psCell, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (psCell)
      {
        ppzStringMemUsage(&(psCell->pzText),pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sTABCELL);
        *pcnBlocks += 1;
      }

    return (psCell);
}


/*-----------------------------------------------
 * Set cell parameters
 *-----------------------------------------------
 */

tVOID vTabCellSetText (tpsTABCELL psCell, tPSZc pzText)
{
    ppzStringCopy(&(psCell->pzText),pzText);
}

/*---------------------------------------------*/

tVOID vTabCellSetSpan (tpsTABCELL psCell, tCOUNT cnColSpan, tCOUNT cnRowSpan)
{
    psCell->cnColSpan = cnColSpan;
    psCell->cnRowSpan = cnRowSpan;
}

/*---------------------------------------------*/

tVOID vTabCellSetColor (tpsTABCELL psCell, tULONG ulColorBack, tULONG ulColorFont)
{
    psCell->ulColorBack = ulColorBack;
    psCell->ulColorFont = ulColorFont;
}

/*---------------------------------------------*/

tVOID vTabCellSetAlign (tpsTABCELL psCell, tCHAR cAlign)
{
    switch (cAlign)
      {
      case 'l' :
      case 'L' :
        psCell->cAlign = 'l';
        break;
      case 'r' :
      case 'R' :
        psCell->cAlign = 'r';
        break;
      default :
        psCell->cAlign = 'c';
      }
}

/*---------------------------------------------*/

tVOID vTabCellSetStyle (tpsTABCELL psCell, tBOOL bBroad, tBOOL bItalic)
{
    psCell->bBroad  = bBroad;
    psCell->bItalic = bItalic;
}


/*-----------------------------------------------
 * Build HTML string
 *-----------------------------------------------
 * Result is empty if ColSpan or RowSpan == 0.
 *-----------------------------------------------
 */

tpPSZ ppzTabCellHtmlStr (tpPSZ ppzStr, tpsTABCELL psCell)
{
    tPSZ  pzColSpan = NULL,
          pzRowSpan = NULL,
          pzColBack = NULL,
          pzColFont = NULL;
    tBOOL bFont = FALSE;

    if ( psCell->cnColSpan && psCell->cnRowSpan )
      {
        if (psCell->cnColSpan > 1)
          {
            ppzStringULongToStr(&pzColSpan,(tULONG)(psCell->cnColSpan),10);
            ppzStringConcatN(&pzColSpan,3," colspan=\"",pzColSpan,"\"");
          }
        if (psCell->cnRowSpan > 1)
          {
            ppzStringULongToStr(&pzRowSpan,(tULONG)(psCell->cnRowSpan),10);
            ppzStringConcatN(&pzRowSpan,3," rowspan=\"",pzRowSpan,"\"");
          }

        if ( (psCell->ulColorBack & 0xff000000) == 0 )
          {
            ppzStringULongToStr(&pzColBack,psCell->ulColorBack,16);
            ppzStringFormatNumber(&pzColBack,pzColBack,6,0,""," bgcolor=\"#","\"");
          }
        if ( (psCell->ulColorFont & 0xff000000) == 0 )
          {
            ppzStringULongToStr(&pzColFont,psCell->ulColorFont,16);
            ppzStringFormatNumber(&pzColFont,pzColFont,6,0,""," color=\"#","\"");
            bFont = TRUE;
          }

        ppzStringConcatN(ppzStr,20,
            "<td",pzColSpan,pzRowSpan," align=\"",
            ( psCell->cAlign == 'c' ? "center" : "" ),
            ( psCell->cAlign == 'l' ? "left"   : "" ),
            ( psCell->cAlign == 'r' ? "right"  : "" ),
            "\"",pzColBack,">",
            ( bFont ? "<font"   : "" ),
            ( bFont ? pzColFont : "" ),
            ( bFont ? ">"       : "" ),
            ( psCell->bBroad  ? "<b>" : "" ),
            ( psCell->bItalic ? "<i>" : "" ),
            ( cnStringLength(psCell->pzText) > 0 ? psCell->pzText : "&#160;" ),
            ( psCell->bItalic ? "</i>" : "" ),
            ( psCell->bBroad  ? "</b>" : "" ),
            ( bFont ? "</font>" : "" ),
            "</td>"
        );

        ppzStringClear(&pzColSpan);
        ppzStringClear(&pzRowSpan);
        ppzStringClear(&pzColBack);
        ppzStringClear(&pzColFont);
      }
    else
        ppzStringClear(ppzStr);

    return (ppzStr);
}


/*----------------------------------------------------------------------*/
