

/* table.c */


/*------------------------------------------------------------------------
 * Module for HTML table generation
 *------------------------------------------------------------------------
 * This version of table is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#include "table.h"
#include "mymemory.h"
#include "mystring.h"
#include "tabcell.h"

/*----------------------------------------------------------------------*/


struct s_TABLE
{
    tCOUNT      cnCols;
    tCOUNT      cnRows;
    tppsTABCELL ppsCells;

    tULONG      ulColorBack;
    tULONG      ulColorFont;
    tCHAR       cAlign;
    tBOOL       bBroad;
    tBOOL       bItalic;

    tCOUNT      cnLoopCol;
    tCOUNT      cnLoopRow;
    tBOOL       bLoopIncCol;
    tBOOL       bLoopIncRow;
};

typedef struct s_TABLE tsTABLE;

/*---------------------------------------------*/

static tVOID v_Table_Resize (tpsTABLE psTab, tCOUNT cnNewCols, tCOUNT cnNewRows);


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Local: Expand or shrink table size
 *-----------------------------------------------
 * Any sizes including 0 are valid.
 *-----------------------------------------------
 */

static tVOID v_Table_Resize (tpsTABLE psTab, tCOUNT cnNewCols, tCOUNT cnNewRows)
{
    tppsTABCELL ppsNewCells;
    tCOUNT      cnNumC,cnNumR,cnC,cnR,cnTabLine,cnNewLine;

    ppsNewCells = pvMemoryAllocArray(psTABCELL,cnNewCols * cnNewRows);

    cnNumC = ( cnNewCols > psTab->cnCols ? cnNewCols : psTab->cnCols );
    cnNumR = ( cnNewRows > psTab->cnRows ? cnNewRows : psTab->cnRows );

    for (cnR = 0; cnR < cnNumR; cnR++)
      {
        cnTabLine = cnR * psTab->cnCols;
        cnNewLine = cnR * cnNewCols;

        for (cnC = 0; cnC < cnNumC; cnC++)
          {
            if ( (cnC < cnNewCols) && (cnR < cnNewRows) )
              {
                if ( (cnC < psTab->cnCols) && (cnR < psTab->cnRows) )
                    ppsNewCells[cnNewLine + cnC] = psTab->ppsCells[cnTabLine + cnC];
                else
                    ppsNewCells[cnNewLine + cnC] = psTabCellCreate();
              }
            else
                psTabCellDestroy(psTab->ppsCells[cnTabLine + cnC]);
          }
      }

    psTab->cnCols = cnNewCols;
    psTab->cnRows = cnNewRows;

    pvMemoryFree(psTab->ppsCells);
    psTab->ppsCells = ppsNewCells;
}


/*-----------------------------------------------
 * Create a new TABLE
 *-----------------------------------------------
 */

tpsTABLE psTableCreate (tCOUNT cnCols, tCOUNT cnRows)
{
    tpsTABLE psTab;

    psTab = pvMemoryAllocObject(sTABLE);

    psTab->cnCols   =
    psTab->cnRows   = 0;
    psTab->ppsCells = NULL;
    v_Table_Resize(psTab,cnCols,cnRows);

    psTab->ulColorBack =
    psTab->ulColorFont = 0xff000000;
    psTab->cAlign      = 'c';
    psTab->bBroad      =
    psTab->bItalic     = FALSE;

    psTab->cnLoopCol   =
    psTab->cnLoopRow   = 0;
    psTab->bLoopIncCol =
    psTab->bLoopIncRow = FALSE;

    return (psTab);
}


/*-----------------------------------------------
 * Destroy a TABLE
 *-----------------------------------------------
 */

tpsTABLE psTableDestroy (tpsTABLE psTab)
{
    tCOUNT cnI;

    if (psTab)
      {
        cnI = psTab->cnCols * psTab->cnRows;
        while (cnI--)
            psTabCellDestroy(psTab->ppsCells[cnI]);
        pvMemoryFree(psTab->ppsCells);

        pvMemoryFree(psTab);
      }

    return (NULL);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsTABLE psTableMemUsage (tpsTABLE psTab, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    tCOUNT cnN,cnI;

    if (psTab)
      {
        cnN = psTab->cnCols * psTab->cnRows;
        for (cnI = 0; cnI < cnN; cnI++)
            psTabCellMemUsage(psTab->ppsCells[cnI],pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sTABLE) +
                      cnMemorySizeofArray(psTABCELL,cnN);
        *pcnBlocks += 1 + ( cnN ? 1 : 0 );
      }

    return (psTab);
}


/*-----------------------------------------------
 * Set loop parameters
 *-----------------------------------------------
 */


tVOID vTableSetColor (tpsTABLE psTab, tULONG ulColorBack, tULONG ulColorFont)
{
    psTab->ulColorBack = ulColorBack;
    psTab->ulColorFont = ulColorFont;
}

/*---------------------------------------------*/

tVOID vTableSetAlign (tpsTABLE psTab, tCHAR cAlign)
{
    psTab->cAlign = cAlign;
}

/*---------------------------------------------*/

tVOID vTableSetStyle (tpsTABLE psTab, tBOOL bBroad, tBOOL bItalic)
{
    psTab->bBroad  = bBroad;
    psTab->bItalic = bItalic;
}


/*-----------------------------------------------
 * Set position parameters
 *-----------------------------------------------
 */


tVOID vTableSetPosition (tpsTABLE psTab, tCOUNT cnCol, tCOUNT cnRow, tBOOL bIncCol, tBOOL bIncRow)
{
    psTab->cnLoopCol   = cnCol;
    psTab->cnLoopRow   = cnRow;
    psTab->bLoopIncCol = bIncCol;
    psTab->bLoopIncRow = bIncRow;
}

/*---------------------------------------------*/

tVOID vTableSetNextRow (tpsTABLE psTab, tCOUNT cnCol)
{
    psTab->cnLoopCol = cnCol;
    psTab->cnLoopRow++;
}

/*---------------------------------------------*/

tVOID vTableSetNextCol (tpsTABLE psTab, tCOUNT cnRow)
{
    psTab->cnLoopCol++;
    psTab->cnLoopRow = cnRow;
}


/*-----------------------------------------------
 * Set next cell Text
 *-----------------------------------------------
 * Uses the actual loop parameters. When the position
 * is out of the actual table dimensions, the table
 * is extended.
 *-----------------------------------------------
 */

tVOID vTableSetCellText (tpsTABLE psTab, tPSZc pzText)
{
    tpsTABCELL psCell;

    if ( (psTab->cnLoopCol >= psTab->cnCols) || (psTab->cnLoopRow >= psTab->cnRows) )
        v_Table_Resize(psTab,
            ( psTab->cnLoopCol >= psTab->cnCols ? psTab->cnLoopCol + 1 : psTab->cnCols ),
            ( psTab->cnLoopRow >= psTab->cnRows ? psTab->cnLoopRow + 1 : psTab->cnRows )
        );

    psCell = psTab->ppsCells[psTab->cnLoopRow * psTab->cnCols + psTab->cnLoopCol];

    vTabCellSetText (psCell,pzText);
    vTabCellSetColor(psCell,psTab->ulColorBack,psTab->ulColorFont);
    vTabCellSetAlign(psCell,psTab->cAlign);
    vTabCellSetStyle(psCell,psTab->bBroad,psTab->bItalic);

    if (psTab->bLoopIncCol) psTab->cnLoopCol++;
    if (psTab->bLoopIncRow) psTab->cnLoopRow++;
}


/*-----------------------------------------------
 * Extend row over all columns
 *-----------------------------------------------
 */

tVOID vTableExtendRow (tpsTABLE psTab, tCOUNT cnRow)
{
    tCOUNT cnRowStart,cnC;

    if (psTab->cnCols)
      {
        if (cnRow >= psTab->cnRows)
            v_Table_Resize(psTab,psTab->cnCols,cnRow + 1);

        cnRowStart = cnRow * psTab->cnCols;
        vTabCellSetSpan(psTab->ppsCells[cnRowStart],psTab->cnCols,1);
        for (cnC = 1; cnC < psTab->cnCols; cnC++)
            vTabCellSetSpan(psTab->ppsCells[cnRowStart + cnC],0,0);
      }
}


/*-----------------------------------------------
 * Build HTML string
 *-----------------------------------------------
 * Quite inefficient with string allocations, copys
 * and concats over and over again, but it works.
 * And it allows a linearized structure, parameters
 * are used on their positions.
 * printf: printf("<td%s>%s</td>",pzAlign,pzText);
 * string: ppzStringConcatN(ppzDst,6,*ppzDst,"<td",pzAlign,">",pzText,"</td>");
 * Additional, Concat takes NULL as an empty string
 * while printf needs something like ( pzStr ? pzStr : "" ).
 *-----------------------------------------------
 */

tpPSZ ppzTableHtmlStr (tpPSZ ppzStr, tpsTABLE psTab)
{
    tPSZ        pzTR = NULL,
                pzTD = NULL;
    tCOUNT      cnC,cnR;
    tppsTABCELL ppsCell;

    ppzStringClear(ppzStr);

    ppsCell = psTab->ppsCells;
    for (cnR = 0; cnR < psTab->cnRows; cnR++)
      {
        for (cnC = 0; cnC < psTab->cnCols; cnC++, ppsCell++)
          {
            ppzTabCellHtmlStr(&pzTD,*ppsCell);
            if (pzTD)
                ppzStringConcatN(&pzTR,4,pzTR,"    ",pzTD,"\n");
          }

        if (pzTR)
          {
            ppzStringConcatN(ppzStr,4,*ppzStr,"  <tr>\n",pzTR,"  </tr>\n");
            ppzStringClear(&pzTR);
          }
      }

    if (*ppzStr)
        ppzStringConcatN(ppzStr,3,
            "<div align=\"center\"><table border=\"1\" cellspacing=\"0\" cellpadding=\"2\">\n",
            *ppzStr,
            "</table></div><br>\n"
        );

    ppzStringClear(&pzTD);

    return (ppzStr);
}


/*----------------------------------------------------------------------*/
