

/* wordcnt.c */


/*------------------------------------------------------------------------
 * Module for big integer arithmetics based on arrays of UWORDs
 *------------------------------------------------------------------------
 * This version of wordcnt is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#include "wordcnt.h"
#include "mymemory.h"
#include "mystring.h"

#include <assert.h>

/*----------------------------------------------------------------------*/


static tVOID v_WordCnt_Resize   (tpsWORDCNT psCnt, tCOUNT cnNewWords);
static tVOID v_WordCnt_AddArray (tpsWORDCNT psCnt, tCOUNT cnIndex, tpcUWORD puwAdd, tCOUNT cnWords, tUWORD uwFactor);


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Initialize a static WORDCNT
 *-----------------------------------------------
 * 0 is valid for cnCounters.
 * The counters are initialized to 32 bit.
 *-----------------------------------------------
 */

tpsWORDCNT psWordCntInit (tpsWORDCNT psCnt, tCOUNT cnCounters)
{
    if (psCnt)
      {
        psCnt->cnWords    = 2;
        psCnt->cnCounters = cnCounters;
        psCnt->puwInts    = pvMemoryAllocArray(UWORD,2 * cnCounters);
      }

    return (psCnt);
}


/*-----------------------------------------------
 * Exit a static WORDCNT
 *-----------------------------------------------
 */

tpsWORDCNT psWordCntExit (tpsWORDCNT psCnt)
{
    if (psCnt)
        pvMemoryFree(psCnt->puwInts);

    return (psCnt);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsWORDCNT psWordCntMemInner (tpsWORDCNT psCnt, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (psCnt)
      {
        *pcnSizeof += cnMemorySizeofArray(UWORD,psCnt->cnWords * psCnt->cnCounters);
        *pcnBlocks += ( psCnt->puwInts ? 1 : 0 );
      }

    return (psCnt);
}


/*-----------------------------------------------
 * Create a new WORDCNT
 *-----------------------------------------------
 * 0 is valid for cnCounters.
 * The counters are initialized to 32 bit.
 *-----------------------------------------------
 */

tpsWORDCNT psWordCntCreate (tCOUNT cnCounters)
{
    return ( psWordCntInit( pvMemoryAllocObject(sWORDCNT) ,cnCounters) );
}


/*-----------------------------------------------
 * Destroy a WORDCNT
 *-----------------------------------------------
 */

tpsWORDCNT psWordCntDestroy (tpsWORDCNT psCnt)
{
    if (psCnt)
      {
        psWordCntExit(psCnt);
        pvMemoryFree(psCnt);
      }

    return (NULL);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsWORDCNT psWordCntMemUsage (tpsWORDCNT psCnt, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (psCnt)
      {
        psWordCntMemInner(psCnt,pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sWORDCNT);
        *pcnBlocks += 1;
      }

    return (psCnt);
}


/*-----------------------------------------------
 * Access macros
 *-----------------------------------------------
 */

#define puw_WordCnt_Counter(psCnt,cnIndex) ( &((psCnt)->puwInts[ (cnIndex) * (psCnt)->cnWords ]) )


/*-----------------------------------------------
 * Init counter with 0
 *-----------------------------------------------
 */

tVOID vWordCntInit0 (tpsWORDCNT psCnt, tCOUNT cnIndex)
{
    tpUWORD puwInt;
    tCOUNT  cnI;

    assert(cnIndex < psCnt->cnCounters);

    puwInt = puw_WordCnt_Counter(psCnt,cnIndex);
    cnI    = psCnt->cnWords;
    while (cnI--)
        *(puwInt++) = 0;
}


/*-----------------------------------------------
 * Init counter with 1
 *-----------------------------------------------
 */

tVOID vWordCntInit1 (tpsWORDCNT psCnt, tCOUNT cnIndex)
{
    tpUWORD puwInt;
    tCOUNT  cnI;

    assert(cnIndex < psCnt->cnCounters);

    puwInt = puw_WordCnt_Counter(psCnt,cnIndex);
    cnI    = psCnt->cnWords;
    if (cnI--)
      {
        *(puwInt++) = 1;

        while (cnI--)
            *(puwInt++) = 0;
      }
}


/*-----------------------------------------------
 * Init all counters with 0
 *-----------------------------------------------
 */

tVOID vWordCntInit0All (tpsWORDCNT psCnt)
{
    tpUWORD puwInt;
    tCOUNT  cnI;

    puwInt = psCnt->puwInts;
    cnI    = psCnt->cnCounters * psCnt->cnWords;
    while (cnI--)
        *(puwInt++) = 0;
}


/*-----------------------------------------------
 * Check if counter == 0
 *-----------------------------------------------
 */

tBOOL bWordCntIs0 (tpsWORDCNT psCnt, tCOUNT cnIndex)
{
    tpUWORD puwInt;
    tCOUNT  cnI;
    tBOOL   bIs0 = TRUE;

    assert(cnIndex < psCnt->cnCounters);

    puwInt = puw_WordCnt_Counter(psCnt,cnIndex);
    cnI    = psCnt->cnWords;
    while (cnI--)
        if ( *(puwInt++) )
          {
            bIs0 = FALSE;
            break;
          }

    return (bIs0);
}


/*-----------------------------------------------
 * Check the used size
 *-----------------------------------------------
 */

tCOUNT cnWordCntUsed (tpsWORDCNT psCnt, tCOUNT cnIndex)
{
    tpUWORD puwInt;
    tCOUNT  cnUsed;

    assert(cnIndex < psCnt->cnCounters);

    cnUsed = psCnt->cnWords;
    puwInt = puw_WordCnt_Counter(psCnt,cnIndex) + cnUsed;

    while ( cnUsed && (*(--puwInt) == 0) )
        cnUsed--;

    return (cnUsed);
}


/*-----------------------------------------------
 * Local: Resize the counters
 *-----------------------------------------------
 * Works in both directions (extend & shrink).
 * On shrinking, the modulus is used.
 * 0 is valid for cnNewWords.
 *-----------------------------------------------
 */

static tVOID v_WordCnt_Resize (tpsWORDCNT psCnt, tCOUNT cnNewWords)
{
    tpUWORD puwNewInts,puwIOld,puwINew;
    tCOUNT  cnI,cnOld,cnNew;

    if (cnNewWords != psCnt->cnWords)
      {
        puwNewInts = pvMemoryAllocArray(UWORD,cnNewWords * psCnt->cnCounters);

        puwIOld = psCnt->puwInts;
        puwINew = puwNewInts;
        for (cnI = 0; cnI < psCnt->cnCounters; cnI++)
          {
            cnOld = psCnt->cnWords;
            cnNew = cnNewWords;

            while (cnOld && cnNew)
              {
                *(puwINew++) = *(puwIOld++);
                cnOld--;
                cnNew--;
              }

            while (cnNew--)
                *(puwINew++) = 0;

            puwIOld += cnOld;
          }

        pvMemoryFree(psCnt->puwInts);
        psCnt->cnWords = cnNewWords;
        psCnt->puwInts = puwNewInts;
      }
}


/*-----------------------------------------------
 * Copy counters
 *-----------------------------------------------
 * Destination is resized if Src doesn't fit
 *-----------------------------------------------
 */

tVOID vWordCntCopy (tpsWORDCNT psCnt, tCOUNT cnICnt, tpsWORDCNT psSrc, tCOUNT cnISrc)
{
    tpUWORD puwCnt,puwSrc;
    tCOUNT  cnCnt,cnSrc;

    assert(cnICnt < psCnt->cnCounters);
    assert(cnISrc < psSrc->cnCounters);

    if (psCnt->cnWords < psSrc->cnWords)
        v_WordCnt_Resize(psCnt,psSrc->cnWords);

    puwCnt = puw_WordCnt_Counter(psCnt,cnICnt);
    puwSrc = puw_WordCnt_Counter(psSrc,cnISrc);
    cnCnt  = psCnt->cnWords;
    cnSrc  = psSrc->cnWords;

    while (cnCnt && cnSrc)
      {
        *(puwCnt++) = *(puwSrc++);
        cnCnt--;
        cnSrc--;
      }

    while (cnCnt--)
        *(puwCnt++) = 0;
}


/*-----------------------------------------------
 * Local: Add UWORD-Array to counter
 *-----------------------------------------------
 * Destination is resized on overflow.
 * Add is added uwFactor times to the counter.
 *
 * ulAdd += (tULONG)(*puwAdd) * uwFactor + *puwInt;
 * is overflow secure in ULONG:
 * 0x0000ffff + 0xffff * 0xffff + 0xffff = 0xffffffff
 *
 * Words(Add) - Words(Cnt) == N (> 0), all UWORDS == 0xffff:
 * ulAdd += (tULONG)(*puwAdd) * uwFactor;
 * gives 0x0000ffff + 0xffff * 0xffff = 0xffff0000
 * ==> cnTmp max == N + 1
 *     auwTmp == 1*0xffff : N*0x0000
 *
 * Thus, auwTmp is limited to 10 UWORDS;
 * when Words(Add) - Words(Cnt) > 9 ==> Resize(Cnt,Words(Add)).
 * This won't happen for standard solitaire.
 *-----------------------------------------------
 */

static tVOID v_WordCnt_AddArray (tpsWORDCNT psCnt, tCOUNT cnIndex, tpcUWORD puwAdd, tCOUNT cnWords, tUWORD uwFactor)
{
    tpUWORD puwInt;
    tCOUNT  cnCnt;
    tULONG  ulAdd;
    tUWORD  auwTmp[10];
    tCOUNT  cnTmp;

    assert(cnIndex < psCnt->cnCounters);

    if ( (psCnt->cnWords + 9) < cnWords )
        v_WordCnt_Resize(psCnt,cnWords);

    puwInt = puw_WordCnt_Counter(psCnt,cnIndex);
    cnCnt  = psCnt->cnWords;
    ulAdd  = 0;
    while (cnCnt && cnWords)
      {
        ulAdd += (tULONG)(*puwAdd) * uwFactor + *puwInt;
        *puwInt = ulAdd & 0xffff;
        ulAdd >>= 16;

        puwInt++;
        puwAdd++;
        cnCnt--;
        cnWords--;
      }

    /*--- This loop executes only when cnWords == 0 ---*/

    while (cnCnt && ulAdd)
      {
        ulAdd += (tULONG)(*puwInt);
        *puwInt = ulAdd & 0xffff;
        ulAdd >>= 16;

        puwInt++;
        cnCnt--;
      }

    /*--- Overflow calculation ---*/

    if (cnWords || ulAdd)
      {
        cnTmp = 0;

        while (cnWords)
          {
            ulAdd += (tULONG)(*puwAdd) * uwFactor;
            auwTmp[cnTmp++] = ulAdd & 0xffff;
            ulAdd >>= 16;

            puwAdd++;
            cnWords--;
          }

        if (ulAdd)
            auwTmp[cnTmp++] = ulAdd & 0xffff;

        cnCnt = psCnt->cnWords;
        v_WordCnt_Resize(psCnt,cnCnt + cnTmp);
        puwInt = puw_WordCnt_Counter(psCnt,cnIndex) + cnCnt;
        while (cnTmp--)
            puwInt[cnTmp] = auwTmp[cnTmp];
      }
}


/*-----------------------------------------------
 * Add counter
 *-----------------------------------------------
 */

tVOID vWordCntAdd (tpsWORDCNT psCnt, tCOUNT cnICnt, tpsWORDCNT psAdd, tCOUNT cnIAdd)
{
    assert(cnIAdd < psAdd->cnCounters);

    v_WordCnt_AddArray(psCnt,cnICnt, puw_WordCnt_Counter(psAdd,cnIAdd) ,psAdd->cnWords,1);
}


/*-----------------------------------------------
 * Add counter multiple times
 *-----------------------------------------------
 */

tVOID vWordCntAddMulti (tpsWORDCNT psCnt, tCOUNT cnICnt, tpsWORDCNT psAdd, tCOUNT cnIAdd, tUWORD uwFactor)
{
    assert(cnIAdd < psAdd->cnCounters);

    v_WordCnt_AddArray(psCnt,cnICnt, puw_WordCnt_Counter(psAdd,cnIAdd) ,psAdd->cnWords,uwFactor);
}


/*-----------------------------------------------
 * Add a single UWORD
 *-----------------------------------------------
 */

tVOID vWordCntAddWord (tpsWORDCNT psCnt, tCOUNT cnICnt, tUWORD uwAdd)
{
    v_WordCnt_AddArray(psCnt,cnICnt,&uwAdd,1,1);
}


/*-----------------------------------------------
 * Convert counter to string
 *-----------------------------------------------
 */

tpPSZ ppzWordCntToStr (tpPSZ ppzStr, tpsWORDCNT psCnt, tCOUNT cnIndex, tUWORD uwBase)
{
    assert(cnIndex < psCnt->cnCounters);

    return ( ppzStringUIntToStr(ppzStr, puw_WordCnt_Counter(psCnt,cnIndex) ,psCnt->cnWords,uwBase) );
}


/*----------------------------------------------------------------------*/
